from contextlib import suppress
from dataclasses import dataclass
from re import search
from json import loads

import pandas as pd

from .extras import ConnectionException, NoSuchConnectionException, QueryingException, Results
from .utils import get_connection_data, get_script_content


@dataclass
class DBConnector:
    """
        Create a connection instance to execute queries

        :type secret_id: string
        :param secret_id: Arn from the secret manager
            You must first make sure that you have the permissions on the resource to invoke

        :type dict_connection: string
        :param dict_connection: Dictionary with connection values (host, username, password, database, driver, port)

        :type timeout: int
        :param timeout: Timeout

        :raises:
            ConnectionException: If provided data through direct connection or secret value are wrong to establish connection
            NotImplementedError: If the driver property has an unmapped value ({PostgreSQL}, {MySQL}, {SQL Server #})
    """

    def __init__(self, secret_id: str = None, dict_connection: dict = None, timeout: int = 5):

        condition = secret_id is not None

        condition2 = dict_connection is not None

        if condition and condition2:
            raise Exception(
                "You can only send one argument 'secret_id' or 'dict_connection', and two were received")
        elif condition or condition2:
            if condition:
                _data = get_connection_data(secret_id)
                if isinstance(_data, str):
                    _data = loads(_data)
            else:
                _data = dict_connection
        else:
            raise ValueError(
                "'secret_id' or 'dict_connection must be specified'")

        self.driver = _data['driver']
        self.host = _data['host']
        self.username = _data['username']
        self.password = _data['password']
        self.port = _data['port']
        self.database = _data['database']

        with suppress(KeyError):
            self.instance_name = _data['instance_name']

        if search(self.driver, "{MySQL}"):
            from pymysql import connect, cursors
            try:
                self.connection = connect(
                    host=self.host,
                    port=int(self.port),
                    user=self.username,
                    password=self.password,
                    database=self.database,
                    charset='utf8mb4',
                    cursorclass=cursors.DictCursor,
                    connect_timeout=timeout
                )
            except Exception as e:
                raise ConnectionException(str(e))
        elif search(self.driver, "{PostgreSQL}"):
            from psycopg2 import connect
            from psycopg2.extras import RealDictCursor
            try:
                self.connection = connect(
                    user=self.username,
                    password=self.password,
                    host=self.host,
                    port=self.port,
                    database=self.database,
                    cursor_factory=RealDictCursor,
                    connect_timeout=timeout
                )
            except Exception as e:
                raise ConnectionException(str(e))
        elif 'MSSQL' in self.driver or 'SQL Server' in self.driver:
            import pyodbc

            _server = "{host}{instance_name},{port}".format(
                host=self.host,
                instance_name='\\' + self.instance_name if self.instance_name else '',
                port=self.port
            )

            str_conn = f"""
                        DRIVER={self.driver};
                        SERVER={_server};
                        DATABASE={self.database};
                        UID={self.username};
                        PWD={self.password};
                    """
            try:
                self.connection = pyodbc.connect(str_conn, timeout=timeout)
            except Exception as e:
                raise ConnectionException(str(e))
        else:
            raise NotImplementedError(
                f'It has not been defined how to handle the supplied controller: {self.driver}')

    def execute_query(
        self, bucket: str = None, file: str = None,
        query: str = None,
        params={},
        results=Results.DICT
    ):
        """
            Execute the statement (it can be a '.sql' file or string with the query) with the specified connection.

            :param bucket: Arn from the s3 Bucket
                You must first make sure that you have the permissions on the resource to invoke
            :type bucket: string

            :param file: file name with extension
                You must first make sure that you have the permissions on the resource to invoke
            :type file: strings

            :param query: query string to be executed
            :type query: string

            :param params: parameters to be replaced as statement in query as statement
            :type params: object

            :param results:Types of expected results
            :type results: NeoConnector.extras.Results

            :raises
                ValueError: If bucket is None ad file is None or query is None.
                NoSuchConnectionException: When trying to run a query and the connection does not exist.
                QueryingException: When you try to run a query, but it is invalid.

            :return: Dictionary list with query results
        """

        is_file = not ((bucket is None) and (file is None))
        is_script = not (query is None)

        if is_file or is_script:
            if bucket and file:
                query = get_script_content(bucket, file)
            try:
                query = query.format(*params, **params)

                data = pd.read_sql_query(
                    query, self.connection, coerce_float=False)

                if results is Results.DICT:
                    return data.to_dict(orient='records')
                elif results is Results.JSON:
                    return data.to_json(orient='records')
                else:
                    return data
            except AttributeError as e:
                raise NoSuchConnectionException(str(e))
            except Exception as e:
                raise QueryingException(str(e))
        else:
            raise ValueError('"Bucket" and "file" or "query" must not be null')



from enum import Enum


class ConnectionException(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        return 'There was an error trying to connecting to database, {0} '.format(self.message)


class QueryingException(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        return 'There was an error executing query, {0} '.format(self.message)


class NoSuchConnectionException(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        return 'You must provide at least one parameter. Bucket and file or just the query through the parameters, {0}'\
            .format(self.message)


class SecretManagerException(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        return 'There was an error retrieving the data located in secret, {0} '.format(
            self.message)


class BucketReadFileException(Exception):
    def __init__(self, *args):
        if args:
            self.message = args[0]
        else:
            self.message = None

    def __str__(self):
        return 'There was an error getting the object from bucket, {0} '.format(self.message)


class Results(Enum):
    JSON = 1
    DICT = 2
    DF = 3



#!/usr/bin/env python

from setuptools import setup

# This setup is suitable for "python setup.py develop".

setup(
        name='awsneo',
        version='0.0.1',
        description='Library for python 3 that allows you to connect to different database engines (PostgreSQL, MySQL, MSSQL) through secrets manager (secret values) or direct connection, also allows you to execute queries from scripts in buckets or directly as a single query',
        author='Juan Rueda',
        author_email='Juan.Rueda@btgpactual.com',
        packages=['awsneo'],
        download_url='https://github.com/jprugo/awsneo/archive/refs/tags/0.0.1.tar.gz'
)


from json import loads
from contextlib import suppress
from re import sub, MULTILINE
import boto3
from botocore.exceptions import ClientError
from .extras import SecretManagerException, BucketReadFileException


with suppress(Exception):
    scm_client = boto3.client('secretsmanager', region_name="us-east-1")
    s3_client = boto3.client('s3')


def get_connection_data(secret_id: str) -> dict:
    
    """
    Retrieves the secret value stored through secrets Manager (AWS) with the connection to the different database
    sources.

        :type secret_id: string
        :param secret_id: Arn from the Secrets Manager resource
            You must first make sure that you have the permissions on the resource to invoke

        :raises
                SecretManagerException: If bucket is None ad file is None or query is None.
                Exception: normal exception

        :return: Secrets manager data dumped
    """
    try:
        data = scm_client.get_secret_value(
            SecretId=secret_id
        )
        if isinstance(data, str):
            data = loads(data)
        return data['SecretString']
    except ClientError as e:
        raise SecretManagerException(str(e))
    except Exception as e:
        raise e


def get_script_content(bucket: str, file_name: str):
    """
        Retrieves the script stored in an s3 bucket (AWS)

        :param bucket: Bucket name from which the script will be retrieved
        :type bucket: string

        :type file_name: string
        :param file_name: file name with extension
            You must first make sure that you have the permissions on the resource to invoke
                :py:meth:`file_name.sql`.

        :raises
                BucketReadFileException: If bucket is None ad file is None or query is None.
                Exception: normal exception

        :return: Script file content
    """
    try:
        bucket = sub(r".*:(.*?)", '', bucket, 0, MULTILINE)
        data = s3_client.get_object(Bucket=bucket, Key=file_name)
        return data['Body'].read().decode('utf-8')
    except ClientError as e:
        raise BucketReadFileException(str(e))
    except Exception as e:
        raise e
